

using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Runtime.Serialization;

namespace EasyImages
{

    

    /// <summary>
    ///  A class used for transporting an image over the wire as a JPEG.
    /// </summary>
    [Serializable]
    public class JPEG
    {

        byte[] payload;

        /// <summary>
        ///  Converts the image to a JPEG
        /// </summary>
        /// <param name="i">Image to be converted</param>
        public JPEG(Image i)
        {
            MemoryStream ms = new MemoryStream();
            i.Save(ms, System.Drawing.Imaging.ImageFormat.Jpeg);
            payload = ms.GetBuffer();
        }

        /// <summary>
        /// Converts the image to a JPEG
        /// </summary>
        /// <param name="i">The image to be Converted</param>
        /// <param name="quality">The quality to save the image at</param>
        public JPEG(Image i, long quality)
        {
            EncoderParameter qualityParam = new EncoderParameter(System.Drawing.Imaging.Encoder.Quality, (long)quality);
            ImageCodecInfo jpegCodec = this.getEncoderInfo("image/jpeg");
            EncoderParameters encoderParams = new EncoderParameters(1);
            encoderParams.Param[0] = qualityParam;

            MemoryStream ms = new MemoryStream();
            i.Save(ms, jpegCodec, encoderParams);
            payload = ms.GetBuffer();
        }


        /// <summary>
        /// Get the JPEG back as a byte array.
        /// </summary>
        /// <returns>The corresponding byte array</returns>
        public byte[] GetByteArray()
        {
            return payload;
        }

        /// <summary>
        /// Get the JPEG back as a raster image.
        /// </summary>
        /// <returns>The corresponding Image</returns>
        public Image GetImage()
        {
            MemoryStream ms = new MemoryStream(payload);
            Image returnImage = Image.FromStream(ms);
            return returnImage;
        }

        private ImageCodecInfo getEncoderInfo(string mimeType)
        {
            // Get image codecs for all image formats
            ImageCodecInfo[] codecs = ImageCodecInfo.GetImageEncoders();

            // Find the correct image codec
            for (int i = 0; i < codecs.Length; i++)
                if (codecs[i].MimeType == mimeType)
                    return codecs[i];
            return null;
        }
    }
    /// <summary>
    /// A Convolution Matrix for creating various image processing effects.  A number of useful presets are included.
    /// Because of the size of the matrix, for functions like blurs it is often useful to perform the same effect several times in a row to get this desired level of blurring.
    /// </summary>
    class DistortionMethod
    {
        public static DistortionMethod NIGHTMARE = new DistortionMethod(100, 100, 100, 100, -800, 100, 100, 100, 100, 1, 0);
        public static DistortionMethod BLUR = new DistortionMethod(1, 1, 1, 1, 1, 1, 1, 1, 1, 9, 0);
        public static DistortionMethod GAUSSIAN_BLUR = new DistortionMethod(1, 2, 1, 2, 4, 2, 1, 2, 1, 16, 0);
        public static DistortionMethod GAUSSIAN_SMOOTH = new DistortionMethod(16, 26, 16, 26, 41, 26, 16, 26, 16, 209, 0);
        public static DistortionMethod SHARPEN_MORE = new DistortionMethod(0, -4, 0, -4, 17, -4, 0, -4, 0, 1, 0);
        public static DistortionMethod SHARPEN = new DistortionMethod(0, -2, 0, -2, 11, -2, 0, -2, 0, 3, 0);
        public static DistortionMethod MEAN_REMOVAL = new DistortionMethod(-1, -1, -1, -1, 9, -1, -1, -1, -1, 1, 0);
        public static DistortionMethod EDGE_DETECT = new DistortionMethod(1, 1, 1, 0, 0, 0, -1, -1, -1, 1, 127);
        public static DistortionMethod EMBOSS_LAPLASCIAN = new DistortionMethod(-1, 0, -1, 0, 4, 0, -1, 0, -1, 1, 127);
        public static DistortionMethod EMBOSS_HORIZONTAL_VERTICAL = new DistortionMethod(0, -1, 0, -1, 4, -1, 0, -1, 0, 1, 127);
        public static DistortionMethod EMBOSS_ALL_DIRECTIONS = new DistortionMethod(-1, -1, -1, -1, 8, -1, -1, -1, -1, 1, 127);
        public static DistortionMethod EMBOSS_LOSSY = new DistortionMethod(1, -2, 1, -2, 4, -2, -2, 1, -2, 1, 127);
        public static DistortionMethod EMBOSS_HORIZONTAL = new DistortionMethod(0, 0, 0, -1, 2, -1, 0, 0, 0, 1, 127);
        public static DistortionMethod EMBOSS_VERTICAL = new DistortionMethod(0, -1, 0, 0, 0, 0, 0, 1, 0, 1, 127);


        public int TopLeft = 0, TopMid = 0, TopRight = 0;
        public int MidLeft = 0, Pixel = 1, MidRight = 0;
        public int BottomLeft = 0, BottomMid = 0, BottomRight = 0;
        public int Factor = 1;
        public int Offset = 0;

        public DistortionMethod()
        {
        }

        public DistortionMethod(int tl, int tm, int tr, int ml, int p, int mr, int bl, int bm, int br, int f, int o)
        {
            this.Pixel = p;
            this.TopLeft = tl;
            this.TopMid = tm;
            this.TopRight = tr;
            this.MidLeft = ml;
            this.MidRight = mr;
            this.BottomLeft = bl;
            this.BottomMid = bm;
            this.BottomRight = br;
            this.Factor = f;
            this.Offset = o;
        }

        /// <summary>
        /// Sets every position in the matrix to some value.
        /// </summary>
        /// <param name="nVal"></param>
        public void SetAll(int nVal)
        {
            TopLeft = TopMid = TopRight = MidLeft = Pixel = MidRight =
                      BottomLeft = BottomMid = BottomRight = nVal;
        }
    }

    public enum DistortionStyle {Blur1 = 1, Blur2 = 2, Blur3 = 3, Blur4 = 4, Blur5 = 5, Blur6 = 6, Blur7 = 7, Sharpen =8, SharpenMore = 9, MeanRemoval= 10, EdgeDetect = 11,
    LaplascianEmboss = 12, EmbossHorizontalVertical = 13, EmbossAllDirections = 14, EmbossLossy = 15, EmbossHorizontal = 16, EmbossVertical = 17};

    public enum DitherStyle { BlackAndWhite, Gray4, Gray16, Color16, Color256, Gray256 };

    public enum OperationType {Subtract};

    /// <summary>
    /// A class used for basic Image Manipulation.  Note that the System.Drwaing.Bitmap class 
    /// contains many useful functions. The functions in this class overwrite the images passed to them.
    /// </summary>
    public abstract class ImageManipulator
    {
        private static int SAT(int v, int m, int n)
        {
            return (((v)<(m))?(m):(((v)>(n))?(n):(v))) ;
        }

        private static int MIN3(int a, int b, int c) 
        {
            return (((a)<(b))?(((c)<(a))?(c):(a)):(((c)<(b))?(c):(b)));
        }

        private static int MAX3(int a, int b, int c)
        { 
            return (((a)>(b))?(((c)>(a))?(c):(a)):(((c)>(b))?(c):(b)));
        }
/*
        /// <summary>
        /// Performs flips on a given image.
        /// </summary>
        /// <param name="flipHorizontal">if true flips horizontal</param>
        /// <param name="flipVertical">if true flips vertical</param>
        /// <param name="im">image to manipulate</param>
        public static void Flip(bool flipHorizontal, bool flipVertical, Bitmap im)
        {
            if (flipHorizontal && flipVertical) im.RotateFlip(RotateFlipType.RotateNoneFlipXY);
            else if (flipVertical) im.RotateFlip(RotateFlipType.RotateNoneFlipY);
            else if (flipHorizontal) im.RotateFlip(RotateFlipType.RotateNoneFlipX);
        }*/

        /// <summary>
        /// Converts an image to a byte array.  Encoding is based on selected ImageFormat.
        /// </summary>
        /// <param name="i">image to convert</param>
        /// <param name="imgf">image format to use</param>
        /// <returns></returns>
        public static byte[] ImageToFormattedByteArray(System.Drawing.Imaging.ImageFormat imgf, Image i)
        {
            MemoryStream ms = new MemoryStream();
            i.Save(ms, imgf);
            return ms.ToArray();
        }

        /*
        public static byte[] ImageToFormattedByteArray(System.Drawing.Imaging.ImageFormat imgf, long quality, Image i)
        {

            EncoderParameter qualityParam = new EncoderParameter(System.Drawing.Imaging.Encoder.Quality, (long)quality);
            ImageCodecInfo thisCodec = ImageManipulator.getEncoderInfo("image/"+imgf.ToString());
            EncoderParameters encoderParams = new EncoderParameters(1);
            encoderParams.Param[0] = qualityParam;
            MemoryStream ms = new MemoryStream();
            i.Save(ms, imgf);
            return ms.ToArray();

        }*/


        private static void DistortBlur(float Clarity, Bitmap bmp)
        {
            for (int i = 0; i < 3; i++)
            {
                Clarity *= 10;
                if (Clarity < 1)
                {
                    Clarity = 1;
                }
                else if (Clarity > 10)
                {
                    Clarity = 10;
                }
                Clarity = (float)(Math.Log10(Clarity));
            }
            int dW = (int)((1 - Clarity) * bmp.Width), dH = (int)((1 - Clarity) * bmp.Height);
            if ((dW <= 1) || (dH <= 1))
            {
                return;
            }

            int[] T = new int[bmp.Width * 3 * bmp.Height];

            int N = dW / 2 + 1; // half-window


            BitmapData data = bmp.LockBits(new Rectangle(0, 0, bmp.Width, bmp.Height),
                                ImageLockMode.ReadWrite,
                                PixelFormat.Format24bppRgb);
            unsafe
            {
                byte* imgPtr = (byte*)(data.Scan0);

                for (int I = 0; I < data.Height; I++)
                {
                    //Console.WriteLine(data.Width + " " + data.Height);
                    int[] S = { 0, 0, 0 };
                    int Z = (I * (data.Stride)) + data.Scan0.ToInt32();
                    byte* prgbSrc = (byte*)Z;

                    for (int J = N + 1; J >= 0; J--)
                    {
                        prgbSrc = (byte*)(Z + (J*3));
                        S[0] += (int)*prgbSrc;
                        prgbSrc++;
                        S[1] += (int)*prgbSrc;
                        prgbSrc++;
                        S[2] += (int)*prgbSrc;
                    }

                    for (int J = N - 1; J > 0; J--)
                    {
                        prgbSrc = (byte*)(Z + (J*3));
                        S[0] += (int)*prgbSrc;
                        prgbSrc++;
                        S[1] += (int)*prgbSrc;
                        prgbSrc++;
                        S[2] += (int)*prgbSrc;
                        prgbSrc++;
                    }

                    for (int J = 0; J < data.Width; J++)
                    {
                        int K1 = J - N - 1; if (K1 < 0) K1 *= -1;
                        int K2 = J + N; if (K2 >= data.Width) K2 = (data.Width * 2) - 1 - K2;
                        prgbSrc = (byte*)(Z + (K1*3));
                        S[0] -= *prgbSrc;
                        prgbSrc++;
                        S[1] -= *prgbSrc;
                        prgbSrc++;
                        S[2] -= *prgbSrc;

                        prgbSrc = (byte*)(Z + (K2*3));
                        S[0] += *prgbSrc;
                        prgbSrc++;
                        S[1] += *prgbSrc;
                        prgbSrc++;
                        S[2] += *prgbSrc;
                        int ZZ = (J * 3) + I * (data.Stride);
                        //int ZZ = (J * 3 * bmp.Height + I * (data.Stride));
                        
                        T[ZZ] = S[0];
                        T[ZZ + 1] = S[1];
                        T[ZZ + 2] = S[2];
                    }


                }

                N = dH / 2 + 1;
                int W = (2 * (dH / 2 + 1) + 1) * (2 * (dW / 2 + 1) + 1);
                for (int I = 0; I < data.Width; I++)
                {
                    int[] S ={ 0, 0, 0 };
                    int Z = (I *3);
                    for (int J = N + 1; J >= 0; J--)
                    {
                        int Z2 = Z + (J*data.Stride);
                        S[0] += T[Z2];
                        S[1] += T[Z2 + 1];
                        S[2] += T[Z2 + 2];
                    }

                    for (int J = N - 1; J > 0; J--)
                    {
                        int Z2 = Z + (J*data.Stride);
                        S[0] += T[Z2];
                        S[1] += T[Z2 + 1];
                        S[2] += T[Z2 + 2];
                    }

                    for (int J = 0; J < data.Height; J++)
                    {
                        int K1 = J - N - 1; if (K1 < 0) K1 *= -1;
                        int Z1 = Z + K1 * (data.Stride);
                        int K2 = J + N; if (K2 >= data.Height) K2 = (data.Height * 2) - 1 - K2;
                        int Z2 = Z + K2 * (data.Stride);
                        S[0] -= T[Z1];
                        S[1] -= T[Z1 + 1];
                        S[2] -= T[Z1 + 2];
                        S[0] += T[Z2];
                        S[1] += T[Z2 + 1];
                        S[2] += T[Z2 + 2];
                        imgPtr = (byte*)data.Scan0;
                        imgPtr += (J * data.Stride) + I*3;
                        *imgPtr = 0;
                       // Console.WriteLine((S[0]/W) + (S[1]/W) + (S[2]/W));
                        *imgPtr = (byte)(S[0]/W);
                        imgPtr++;
                        *imgPtr = (byte)(S[1]/W);
                        imgPtr++;
                        *imgPtr = (byte)(S[2]/W);
                    }


                }


            }
            bmp.UnlockBits(data);
        }

        private static ImageCodecInfo getEncoderInfo(string mimeType)
        {
            // Get image codecs for all image formats
            ImageCodecInfo[] codecs = ImageCodecInfo.GetImageEncoders();

            // Find the correct image codec
            for (int i = 0; i < codecs.Length; i++)
                if (codecs[i].MimeType == mimeType)
                    return codecs[i];
            return null;
        }

        /// <summary>
        /// Converts a byte array back to an image.
        /// </summary>
        /// <param name="b">byte array of the image</param>
        /// <returns>the image corresponding to the byte array</returns>
        public static Image FormattedByteArrayToImage(byte[] b)
        {
            MemoryStream ms = new MemoryStream(b);
            Image returnImage = Image.FromStream(ms);
            return returnImage;
        }

        static int[] ColorReturnValue4=  {0x00,0x80,0xc0,0xff};
        static int[] ColorReturnValue16 = { 0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff };
        static int[,] VGA16Colors=new int[16,3]{{0x00,0x00,0x00},{0x00,0x00,0x80},{0x00,0x00,0xff},
		{0x00,0x80,0x00},{0x00,0x80,0x80},{0x00,0xff,0x00},{0x00,0xff,0xff},{0x80,0x00,0x00},
		{0x80,0x00,0x80},{0x80,0x80,0x00},{0x80,0x80,0x80},{0xc0,0xc0,0xc0},{0xff,0x00,0x00},
		{0xff,0x00,0xff},{0xff,0xff,0x00},{0xff,0xff,0xff}};					
	    static int[,] VGA256Colors=new int[256,3]{{0,0,0},{128,0,0},{0,128,0},{128,128,0},{0,0,128},
		{128,0,128},{0,128,128},{192,192,192},{192,220,192},{212,208,200},{0,0,51},{51,0,0},
		{51,0,51},{0,51,51},{22,22,22},{28,28,28},{34,34,34},{41,41,41},{85,85,85},{77,77,77},
		{66,66,66},{57,57,57},{255,124,128},{255,80,80},{214,0,147},{204,236,255},
		{239,214,198},{231,231,214},{173,169,144},{51,255,0},{102,0,0},{153,0,0},{204,0,0},
		{0,51,0},{51,51,0},{102,51,0},{153,51,0},{204,51,0},{255,51,0},{0,102,0},{51,102,0},
		{102,102,0},{153,102,0},{204,102,0},{255,102,0},{0,153,0},{51,153,0},{102,153,0},
		{153,153,0},{204,153,0},{255,153,0},{0,204,0},{51,204,0},{102,204,0},{153,204,0},
		{204,204,0},{255,204,0},{102,255,0},{153,255,0},{204,255,0},{0,255,51},{51,0,255},
		{102,0,51},{153,0,51},{204,0,51},{255,0,51},{0,51,255},{51,51,51},{102,51,51},
		{153,51,51},{204,51,51},{255,51,51},{0,102,51},{51,102,51},{102,102,51},{153,102,51},
		{204,102,51},{255,102,51},{0,153,51},{51,153,51},{102,153,51},{153,153,51},
		{204,153,51},{255,153,51},{0,204,51},{51,204,51},{102,204,51},{153,204,51},
		{204,204,51},{255,204,51},{51,255,51},{102,255,51},{153,255,51},{204,255,51},
		{255,255,51},{0,0,102},{51,0,102},{102,0,102},{153,0,102},{204,0,102},{255,0,102},
		{0,51,102},{51,51,102},{102,51,102},{153,51,102},{204,51,102},{255,51,102},
		{0,102,102},{51,102,102},{102,102,102},{153,102,102},{204,102,102},{0,153,102},
		{51,153,102},{102,153,102},{153,153,102},{204,153,102},{255,153,102},{0,204,102},
		{51,204,102},{153,204,102},{204,204,102},{255,204,102},{0,255,102},{51,255,102},
		{153,255,102},{204,255,102},{255,0,204},{204,0,255},{0,153,153},{153,51,153},
		{153,0,153},{204,0,153},{0,0,153},{51,51,153},{102,0,153},{204,51,153},{255,0,153},
		{0,102,153},{51,102,153},{102,51,153},{153,102,153},{204,102,153},{255,51,153},
		{51,153,153},{102,153,153},{153,153,153},{204,153,153},{255,153,153},{0,204,153},
		{51,204,153},{102,204,102},{153,204,153},{204,204,153},{255,204,153},{0,255,153},
		{51,255,153},{102,204,153},{153,255,153},{204,255,153},{255,255,153},{0,0,204},
		{51,0,153},{102,0,204},{153,0,204},{204,0,204},{0,51,153},{51,51,204},{102,51,204},
		{153,51,204},{204,51,204},{255,51,204},{0,102,204},{51,102,204},{102,102,153},
		{153,102,204},{204,102,204},{255,102,153},{0,153,204},{51,153,204},{102,153,204},
		{153,153,204},{204,153,204},{255,153,204},{0,204,204},{51,204,204},{102,204,204},
		{153,204,204},{204,204,204},{255,204,204},{0,255,204},{51,255,204},{102,255,153},
		{153,255,204},{204,255,204},{255,255,204},{51,0,204},{102,0,255},{153,0,255},
		{0,51,204},{51,51,255},{102,51,255},{153,51,255},{204,51,255},{255,51,255},
		{0,102,255},{51,102,255},{102,102,204},{153,102,255},{204,102,255},{255,102,204},
		{0,153,255},{51,153,255},{102,153,255},{153,153,255},{204,153,255},{255,153,255},
		{0,204,255},{51,204,255},{102,204,255},{153,204,255},{204,204,255},{255,204,255},
		{51,255,255},{102,255,204},{153,255,255},{204,255,255},{255,102,102},{102,255,102},
		{255,255,102},{102,102,255},{255,102,255},{102,255,255},{165,0,33},{95,95,95},
		{119,119,119},{134,134,134},{150,150,150},{203,203,203},{178,178,178},{215,215,215},
		{221,221,221},{227,227,227},{234,234,234},{241,241,241},{248,248,248},{255,251,240},
		{58,110,165},{128,128,128},{255,0,0},{0,255,0},{255,255,0},{0,0,255},{255,0,255},
		{0,255,255},{255,255,255}};


        private static int DitherCalculateDistance(int rvalue, int bvalue, int gvalue, int red, int green, int blue) 
        {
            return ((bvalue - blue) * (bvalue - blue)) + ((gvalue - green) * (gvalue - green)) + ((rvalue - red) * (rvalue - red));
        }
        
        static int DitherGetHue(int red, int green, int blue)
        {
            int n = MAX3(blue, green, red);
            int h = n - MIN3(blue, green, red);
            if ((h != 0) && (n > 26) && (n < 229))
            {
                if (red == n)
                {
                    h = 42 * (green - blue) / h;
                }
                else if (green == n)
                {
                    h = 85 + 42 * (blue - red) / h;
                }
                else
                {
                    h = 170 + 42 * (red - green) / h;
                }
                if (h < 0)
                {
                    h += 255;
                }
            }
            else
            {
                h = 0;
            }
            return h;
        }

        private static int[] DitherColorValue(int color, int rvalue, int gvalue, int bvalue)
        {

    	    int CurrentDistance;
            int SmallestDistanceSquared;
            int SmallestPosition;
            int[] ReturnRGBTriple = new int[3]; 
            int[] tempRGBTriple = new int[3];
            int[] ReturnErrorTriple= new int[3];
            //int red = 255;
            //int green = 255;
            //int blue = 255;
            switch(color)
            {
                default:
	                return ReturnErrorTriple;
                case 16:
	                SmallestDistanceSquared = DitherCalculateDistance(rvalue, bvalue, gvalue, VGA16Colors[0,0], VGA16Colors[0,1], VGA16Colors[0,2]) + 4 * DitherGetHue(rvalue, gvalue, bvalue);
	                SmallestPosition=0;
	                break;
                case 256:
                    SmallestDistanceSquared = DitherCalculateDistance(rvalue, bvalue, gvalue, VGA256Colors[0,0], VGA256Colors[0,1], VGA256Colors[0,2]) + 4 * DitherGetHue(rvalue, gvalue, bvalue);
	                SmallestPosition=0;
	                break;
	        }

            for (int count = 0; count < color; count++)
            {
                switch (color)
                {
                        
                    default:
                        return ReturnErrorTriple;
                    case 16:
                        CurrentDistance = DitherCalculateDistance(rvalue, bvalue, gvalue, VGA16Colors[0,0], VGA16Colors[0,1], VGA16Colors[0,2]);
                        tempRGBTriple[0] = VGA16Colors[count,0];
                        tempRGBTriple[1] = VGA16Colors[count,1];
                        tempRGBTriple[2] = VGA16Colors[count,2];
                        break;
                    case 256:
                        CurrentDistance = DitherCalculateDistance(rvalue, bvalue, gvalue, VGA256Colors[count, 0], VGA256Colors[count, 1], VGA256Colors[count, 2]);
                        tempRGBTriple[0] = VGA256Colors[count,0];
                        tempRGBTriple[1] = VGA256Colors[count,1];
                        tempRGBTriple[2] = VGA256Colors[count,2];
                        break;
                }
                CurrentDistance = CurrentDistance + 4 * DitherGetHue(tempRGBTriple[0], tempRGBTriple[1], tempRGBTriple[2]);
                if (CurrentDistance < SmallestDistanceSquared)
                {
                    SmallestDistanceSquared = CurrentDistance;
                    SmallestPosition = count;
                    
                }
            }
            switch (color)
            {
                default:
                    return ReturnErrorTriple;
                case 16:
                    ReturnErrorTriple[0] = VGA16Colors[SmallestPosition,0];
                    ReturnErrorTriple[1] = VGA16Colors[SmallestPosition,1];
                    ReturnErrorTriple[2] = VGA16Colors[SmallestPosition,2];
                    break;
                case 256:
                    ReturnErrorTriple[0] = VGA256Colors[SmallestPosition,0];
                    ReturnErrorTriple[1] = VGA256Colors[SmallestPosition,1];
                    ReturnErrorTriple[2] = VGA256Colors[SmallestPosition,2];
                    break;
            }
            return ReturnRGBTriple;
        }

        private static int DitherColor(int Color, int Yvalue)
        {
            int count;
            int returnvalue = 0;
	        int comparevalue;

	    	
	        switch(Color)
	        {
	            case 2:
		            returnvalue=(Yvalue<128)?0:0xff;
		            break;
	            case 4:
		            for(count=0; count<4; count++)
		            {
			            comparevalue=count*64+64;
			            if(Yvalue<comparevalue)
			            {
				            return ColorReturnValue4[count];
			            }
                    }
		            break;
	            case 16:
		            for(count=0; count<16; count++)
		            {
			            comparevalue=count*16+16;
			            if(Yvalue<comparevalue)
			            {
				            return ColorReturnValue16[count];
			            }
		            }
		            break;
	            case 255:
		            //TODO: Add implementation for 256 Colors here
		            break;
	            default:
		            returnvalue=0;
		            break;
	            }
	            return returnvalue;
        }

        public static void Dither(DitherStyle ds, Bitmap im)
        {
            Bitmap bm = im as Bitmap;
            int Colors = 0;
            bool Grayscale = false;
            switch (ds)
            {

                case DitherStyle.BlackAndWhite:
                    Colors = 2;
                    Grayscale = true;
                    break;
                case DitherStyle.Gray4:
                    Colors = 4;
                    Grayscale = true;
                    break;
                case DitherStyle.Gray16:
                    Colors = 16;
                    Grayscale = true;
                    break;
                case DitherStyle.Color16:
                    Colors = 16;
                    Grayscale = false;
                    break;
                case DitherStyle.Color256:
                    Colors = 256;
                    Grayscale = false;
                    break;
                case DitherStyle.Gray256:
                    Colors = 256;
                    Grayscale = true;
                    break;
            }

            BitmapData data = bm.LockBits(new Rectangle(0, 0, bm.Width, bm.Height),
                    ImageLockMode.ReadWrite,
                    PixelFormat.Format24bppRgb);

            if (Grayscale)
            {
                if (256 == Colors)
                {

                    unsafe
                    {
                        byte* imgPtr = (byte*)(data.Scan0);
                        for (int i = 0; i < data.Height; i++)
                        {
                            for (int j = 0; j < data.Width; j++)
                            {
                                int b;
                                int g;
                                int r;
                                int avg;
                                byte* smptr = imgPtr;
                                b = *smptr;
                                smptr++;
                                g = *smptr;
                                smptr++;
                                r = *smptr;

                                avg = (b + g + r) / 3;

                                smptr = imgPtr;
                                *smptr = (byte)avg;
                                smptr++;
                                *smptr = (byte)avg;
                                smptr++;
                                *smptr = (byte)avg; ;

                                imgPtr += 3;
                            }
                            imgPtr += data.Stride - data.Width * 3;
                        }
                    }
                    bm.UnlockBits(data);
                    return;
                }
                else
                {
                    unsafe
                    {
                        byte* imgPtr = (byte*)(data.Scan0);
                        for (int i = 0; i < data.Height; i++)
                        {
                            for (int j = 0; j < data.Width; j++)
                            {
                                int b;
                                int g;
                                int r;
                                int avg;
                                byte* smptr = imgPtr;
                                b = *smptr;
                                smptr++;
                                g = *smptr;
                                smptr++;
                                r = *smptr;

                                avg = (b + g + r) / 3;

                                smptr = imgPtr;
                                *smptr = (byte)avg;
                                smptr++;
                                *smptr = (byte)avg;
                                smptr++;
                                *smptr = (byte)avg; ;

                                imgPtr += 3;
                            }
                            imgPtr += data.Stride - data.Width * 3;
                        }

                        imgPtr = (byte*)(data.Scan0);

                        for (int i = 0; i < (data.Height - 1); i++)
                        {
                            for (int j = 0; j < data.Width; j++)
                            {
                                byte* smptr = imgPtr;
                                int b = *smptr;
                                int yvalue = DitherColor(Colors, b);
                                int error = b - yvalue;
                                int errorDiffusion = 0;
                                smptr = imgPtr;
                                *smptr = (byte)yvalue;
                                smptr++;
                                *smptr = (byte)yvalue;
                                smptr++;
                                *smptr = (byte)yvalue;

                                
                                smptr = imgPtr +3 ;
                                b = *smptr;
                                errorDiffusion = (16 * b + 7 * error) / 16;
                                errorDiffusion = SAT(errorDiffusion, 0, 255);
                                *smptr = (byte)errorDiffusion;
                                smptr++;
                                *smptr = (byte)errorDiffusion;
                                smptr++;
                                *smptr = (byte)errorDiffusion;

                                if (j > 0)
                                {
                                    smptr = imgPtr + (data.Stride - data.Width * 3) + -3;
                                    b = *smptr;
                                    errorDiffusion = (16 * b + 3 * error) / 16;
                                    errorDiffusion = SAT(errorDiffusion, 0, 255);
                                    *smptr = (byte)errorDiffusion;
                                    smptr++;
                                    *smptr = (byte)errorDiffusion;
                                    smptr++;
                                    *smptr = (byte)errorDiffusion;
                                }

                                smptr = imgPtr + (data.Stride - data.Width * 3);
                                b = *smptr;
                                errorDiffusion = (16 * b + 5 * error) / 16;
                                errorDiffusion = SAT(errorDiffusion, 0, 255);
                                *smptr = (byte)errorDiffusion;
                                smptr++;
                                *smptr = (byte)errorDiffusion;
                                smptr++;
                                *smptr = (byte)errorDiffusion;

                                if (j < (data.Width - 1))
                                {
                                    smptr = imgPtr + (data.Stride - data.Width * 3) + 3;
                                    b = *smptr;
                                    errorDiffusion = (16 * b + error) / 16;
                                    errorDiffusion = SAT(errorDiffusion, 0, 255);
                                    *smptr = (byte)errorDiffusion;
                                    smptr++;
                                    *smptr = (byte)errorDiffusion;
                                    smptr++;
                                    *smptr = (byte)errorDiffusion;
                                } 

                                imgPtr += 3;
                            }
                            imgPtr += data.Stride - data.Width * 3;
                        }
                    }
                    bm.UnlockBits(data);
                    return;
                }
            }
            else
            {

                unsafe
                {
                    byte* imgPtr = (byte*)(data.Scan0);

                    for (int i = 0; i < (data.Height - 1); i++)
                    {
                        for (int j = 0; j < data.Width; j++)
                        {
                            int b;
                            int g;
                            int r;
                           // int avg;
                            byte* smptr = imgPtr;
                            b = *smptr;
                            smptr++;
                            g = *smptr;
                            smptr++;
                            r = *smptr;
                            int[] result = DitherColorValue(Colors, r, g, b);
                            int rerror = r - result[0];
                            int gerror = g - result[1];
                            int berror = b - result[2];
                            int[] values = DitherSetColor(r, g, b, 7, rerror, gerror, berror);
                            smptr = imgPtr;
                            *smptr = (byte)values[2];
                            smptr++;
                            *smptr = (byte)values[1];
                            smptr++;
                            *smptr = (byte)values[0];

                            if (j > 0)
                            {
                                smptr = imgPtr + (data.Stride - data.Width * 3) + -3;
                                b = *smptr;
                                smptr++;
                                g = *smptr;
                                smptr++;
                                r = *smptr;
                                values = DitherSetColor(r, g, b, 3, rerror, gerror, berror);
                                smptr = imgPtr + (data.Stride - data.Width * 3) + -3;
                                *smptr = (byte)values[2];
                                smptr++;
                                *smptr = (byte)values[1];
                                smptr++;
                                *smptr = (byte)values[0];
                            }
                            if (j < (data.Width - 1))
                            {
                                smptr = imgPtr + (data.Stride - data.Width * 3) + 3;
                                b = *smptr;
                                smptr++;
                                g = *smptr;
                                smptr++;
                                r = *smptr;
                                values = DitherSetColor(r, g, b, 1, rerror, gerror, berror);
                                smptr = imgPtr + (data.Stride - data.Width * 3) + 3;
                                *smptr = (byte)values[2];
                                smptr++;
                                *smptr = (byte)values[1];
                                smptr++;
                                *smptr = (byte)values[0];

                            }



                            imgPtr += 3;
                        }
                        imgPtr += data.Stride - data.Width * 3;
                    }
                    
                }
                bm.UnlockBits(data);
                return;
            }
        }

        private static int[] DitherSetColor(int r, int g, int b, int Ratio, int redError, int greenError, int blueError)
        {
            int[] returnInt = new int[3];
            int tempRed=0, tempBlue=0, tempGreen=0;
            tempRed = Math.Abs(redError);
            tempBlue = Math.Abs(blueError);
            tempGreen = Math.Abs(redError);
            tempRed=(16*r+Ratio*tempRed)/16;
	        tempGreen=(16*g+Ratio*tempGreen)/16;
	        tempBlue=(16*b+Ratio*tempBlue)/16;
            returnInt[0] = SAT(tempRed, 0, 255);
            returnInt[1] = SAT(tempGreen, 0, 255);
            returnInt[2] = SAT(tempBlue, 0, 255);
            return returnInt;

        }

        public static void Operation(OperationType o, ref Bitmap photo1, Bitmap photo2)
        {
            switch (o)
            {
                case OperationType.Subtract:
                photo1 = new Bitmap(ImageManipulator.Subtract(photo1, photo2));
                    break;
            }
        }

        /// <summary>
        /// AlphaBlends between two images, replacing the first.
        /// The value sets the blending. The value must be between 0 and 1. .7 would result in the resulting image being weighted 70% in favor of image one and 30% in image two.
        /// </summary>
        /// <param name="photo1">The first image.  The result will overwrite this one.</param>
        /// <param name="photo2">The second image.</param>
        /// <param name="blendAmount">A value between 0 and 1 that defines the weight of the blending.  </param>
        public static void AlphaBlend(ref Bitmap photo1, Bitmap photo2, float blendAmount)
        {
            if (photo1 == null | photo2 == null)
            {
                return;
            }
            if ((photo1.Width != photo2.Width) || (photo1.Height != photo2.Height))
            {
                return;
            }

            Bitmap bm1 = photo1 as Bitmap;
            Bitmap bm2 = photo2 as Bitmap;
            Bitmap output = bm1.Clone() as Bitmap;

            BitmapData data1 = bm1.LockBits(new Rectangle(0, 0, bm1.Width, bm1.Height),
                    ImageLockMode.ReadOnly,
                    PixelFormat.Format24bppRgb);

            BitmapData data2 = bm2.LockBits(new Rectangle(0, 0, bm2.Width, bm2.Height),
                    ImageLockMode.ReadOnly,
                    PixelFormat.Format24bppRgb);

            BitmapData outputData = output.LockBits(new Rectangle(0, 0, output.Width, output.Height),
        ImageLockMode.WriteOnly,
        PixelFormat.Format24bppRgb);

            unsafe
            {
                byte* imgPtr1 = (byte*)(data1.Scan0);
                byte* imgPtr2 = (byte*)(data2.Scan0);
                byte* dataPtr = (byte*) (outputData.Scan0);
                for (int i = 0; i < data1.Height; i++)
                {
                    for (int j = 0; j < data1.Width; j++)
                    {
                        //((*prgbForeground)>(*prgbBackground))?(*prgbForeground)-(*prgbBackground):(*prgbBackground)-(*prgbForeground);

                        *dataPtr = (byte)((((float)*imgPtr1) * blendAmount) + (((float)*imgPtr2) * (1 - blendAmount)));
                        imgPtr1 += 1;
                        imgPtr2 += 1;
                        dataPtr += 1;
                        *dataPtr = (byte)((((float)*imgPtr1) * blendAmount) + (((float)*imgPtr2) * (1 - blendAmount)));
                        imgPtr1 += 1;
                        imgPtr2 += 1;
                        dataPtr += 1;
                        *dataPtr = (byte)((((float)*imgPtr1) * blendAmount) + (((float)*imgPtr2) * (1 - blendAmount)));
                        imgPtr1 += 1;
                        imgPtr2 += 1;
                        dataPtr += 1;
                    }
                    imgPtr1 += data1.Stride - data1.Width * 3;
                    imgPtr2 += data2.Stride - data2.Width * 3;
                    dataPtr += outputData.Stride - outputData.Width * 3;
                }
            }
            bm1.UnlockBits(data1);
            bm2.UnlockBits(data2);
            output.UnlockBits(outputData);
            photo1 = output;
            

        }
        

        private static Bitmap Subtract(Bitmap photo1, Bitmap photo2)
        {
            if (photo1 == null | photo2 == null)
            {
                return null;
            }
            if ((photo1.Width != photo2.Width) || (photo1.Height != photo2.Height))
            {
                return null;
            }

            Bitmap bm1 = photo1 as Bitmap;
            Bitmap bm2 = photo2 as Bitmap;
            Bitmap output = bm1.Clone() as Bitmap;
            //Bitmap output = bm1;

            BitmapData data1 = bm1.LockBits(new Rectangle(0, 0, bm1.Width, bm1.Height),
                    ImageLockMode.ReadOnly,
                    PixelFormat.Format24bppRgb);

            BitmapData data2 = bm2.LockBits(new Rectangle(0, 0, bm2.Width, bm2.Height),
                    ImageLockMode.ReadOnly,
                    PixelFormat.Format24bppRgb);

            BitmapData outputData = output.LockBits(new Rectangle(0, 0, output.Width, output.Height),
                    ImageLockMode.WriteOnly,
                    PixelFormat.Format24bppRgb);
            
            unsafe
            {
                byte* imgPtr1 = (byte*)(data1.Scan0);
                byte* imgPtr2 = (byte*)(data2.Scan0);
                byte* dataPtr = (byte*)(outputData.Scan0);

                for (int i = 0; i < data1.Height; i++)
                {
                    for (int j = 0; j < data1.Width; j++)
                    {
                        //((*prgbForeground)>(*prgbBackground))?(*prgbForeground)-(*prgbBackground):(*prgbBackground)-(*prgbForeground);

                        *dataPtr = ((*imgPtr1) > (*imgPtr2)) ? (byte) ((*imgPtr1) - (*imgPtr2)) : (byte) ((*imgPtr2) - (*imgPtr1));
                        imgPtr1 += 1;
                        imgPtr2 += 1;
                        dataPtr += 1;
                        *dataPtr = ((*imgPtr1) > (*imgPtr2)) ? (byte)((*imgPtr1) - (*imgPtr2)) : (byte) ((*imgPtr2) - (*imgPtr1));
                        imgPtr1 += 1;
                        imgPtr2 += 1;
                        dataPtr += 1;
                        *dataPtr = ((*imgPtr1) > (*imgPtr2)) ? (byte)((*imgPtr1) - (*imgPtr2)) : (byte) ((*imgPtr2) - (*imgPtr1));
                        imgPtr1 += 1;
                        imgPtr2 += 1;
                        dataPtr += 1;
                    }
                    imgPtr1 += data1.Stride - data1.Width * 3;
                    imgPtr2 += data2.Stride - data2.Width * 3;
                    //dataPtr += data1.Stride - data1.Width * 3;
                    dataPtr += outputData.Stride - outputData.Width * 3;
                }
            }
            bm1.UnlockBits(data1);
            bm2.UnlockBits(data2);
            output.UnlockBits(outputData);
            return output;
        }

        public static Bitmap BinarySubtract(Bitmap photo1, Bitmap photo2, int threshold, Color pos, Color neg)
        {
            if (photo1 == null | photo2 == null)
            {
                return null;
            }
            if ((photo1.Width != photo2.Width) || (photo1.Height != photo2.Height))
            {
                return null;
            }

            Bitmap bm1 = photo1 as Bitmap;
            Bitmap bm2 = photo2 as Bitmap;
            Bitmap output = bm1.Clone() as Bitmap;
            //Bitmap output = bm1;

            BitmapData data1 = bm1.LockBits(new Rectangle(0, 0, bm1.Width, bm1.Height),
                    ImageLockMode.ReadOnly,
                    PixelFormat.Format24bppRgb);

            BitmapData data2 = bm2.LockBits(new Rectangle(0, 0, bm2.Width, bm2.Height),
                    ImageLockMode.ReadOnly,
                    PixelFormat.Format24bppRgb);

            BitmapData outputData = output.LockBits(new Rectangle(0, 0, output.Width, output.Height),
                    ImageLockMode.WriteOnly,
                    PixelFormat.Format24bppRgb);

            unsafe
            {
                byte* imgPtr1 = (byte*)(data1.Scan0);
                byte* imgPtr2 = (byte*)(data2.Scan0);
                byte* dataPtr = (byte*)(outputData.Scan0);
                byte* rptr, gptr, bptr;

                for (int i = 0; i < data1.Height; i++)
                {
                    for (int j = 0; j < data1.Width; j++)
                    {
                        //((*prgbForeground)>(*prgbBackground))?(*prgbForeground)-(*prgbBackground):(*prgbBackground)-(*prgbForeground);

                        *dataPtr = ((*imgPtr1) > (*imgPtr2)) ? (byte)((*imgPtr1) - (*imgPtr2)) : (byte)((*imgPtr2) - (*imgPtr1));
                        bptr = dataPtr;
                        imgPtr1 += 1;
                        imgPtr2 += 1;
                        dataPtr += 1;
                        *dataPtr = ((*imgPtr1) > (*imgPtr2)) ? (byte)((*imgPtr1) - (*imgPtr2)) : (byte)((*imgPtr2) - (*imgPtr1));
                        gptr = dataPtr;
                        imgPtr1 += 1;
                        imgPtr2 += 1;
                        dataPtr += 1;
                        *dataPtr = ((*imgPtr1) > (*imgPtr2)) ? (byte)((*imgPtr1) - (*imgPtr2)) : (byte)((*imgPtr2) - (*imgPtr1));
                        rptr = dataPtr;
                        imgPtr1 += 1;
                        imgPtr2 += 1;
                        dataPtr += 1;

                        if ((int)(*rptr * 299 / 1000) +
                            (int)(*gptr * 587 / 1000) +
                            (int)(*bptr * 114 / 1000) > threshold)
                        {
                            *rptr = pos.R;
                            *gptr = pos.G;
                            *bptr = pos.B;
                        }
                        else
                        {
                            *rptr = neg.R;
                            *gptr = neg.G;
                            *bptr = neg.B;
                        }
                    }
                    imgPtr1 += data1.Stride - data1.Width * 3;
                    imgPtr2 += data2.Stride - data2.Width * 3;
                    //dataPtr += data1.Stride - data1.Width * 3;
                    dataPtr += outputData.Stride - outputData.Width * 3;
                }
            }
            bm1.UnlockBits(data1);
            bm2.UnlockBits(data2);
            output.UnlockBits(outputData);
            return output;
        }

        public static void Resize(int newWidth, int newHeight, ref Bitmap bmp)
        {
 //           if (!Smooth)
//            {
//            }
//            else
//            {
                Bitmap result = new Bitmap(newWidth, newHeight);
                using (Graphics g = Graphics.FromImage((Image)result))
                    g.DrawImage(bmp, 0, 0, newWidth, newHeight);
                bmp = result;
  //          }
        }


        /// <summary>
        ///  Performs a convolution based distortion on the image.
        /// Several defaults are available in the ConvMatrix class.
        /// </summary>
        /// <param name="im">image to process</param>
        /// <param name="distortion">distortion to use</param>
        public static void Distort(DistortionStyle ds, Bitmap im)
        {
            DistortionMethod distortion = null;
            switch (ds)
            {
                case DistortionStyle.Blur1:
                    DistortBlur((float).95, im);
                    break;
                case DistortionStyle.Blur2:
                    DistortBlur((float).80, im);
                    break;
                case DistortionStyle.Blur3:
                    DistortBlur((float).65, im);
                    break;
                case DistortionStyle.Blur4:
                    DistortBlur((float).50, im);
                    break;
                case DistortionStyle.Blur5:
                    DistortBlur((float).35, im);
                    break;
                case DistortionStyle.Blur6:
                    DistortBlur((float).20, im);
                    break;
                case DistortionStyle.Blur7:
                    DistortBlur((float).05, im);
                    break;

                case DistortionStyle.EdgeDetect:
                    distortion = DistortionMethod.EDGE_DETECT;
                    break;
                case DistortionStyle.EmbossAllDirections:
                    distortion = DistortionMethod.EMBOSS_ALL_DIRECTIONS;
                    break;
                case DistortionStyle.EmbossHorizontal:
                    distortion = DistortionMethod.EMBOSS_HORIZONTAL;
                    break;
                case DistortionStyle.EmbossHorizontalVertical:
                    distortion = DistortionMethod.EMBOSS_HORIZONTAL_VERTICAL;
                    break;
                case DistortionStyle.EmbossLossy:
                    distortion = DistortionMethod.EMBOSS_LOSSY;
                    break;
                case DistortionStyle.EmbossVertical:
                    distortion = DistortionMethod.EMBOSS_VERTICAL;
                    break;
                case DistortionStyle.LaplascianEmboss:
                    distortion = DistortionMethod.EMBOSS_LAPLASCIAN;
                    break;
                case DistortionStyle.MeanRemoval:
                    distortion = DistortionMethod.MEAN_REMOVAL;
                    break;
                case DistortionStyle.SharpenMore:
                    distortion = DistortionMethod.SHARPEN_MORE;
                    break;
                case DistortionStyle.Sharpen:
                    distortion = DistortionMethod.SHARPEN;
                    break;
            }
            if(null != distortion){
                Conv3x3(im as Bitmap, distortion);
            }
        }

        /// <summary>
        ///  Distorts the image by the amount specified where a higher value indicates more distortion.
        /// </summary>
        /// <param name="im">image to process</param>
        /// <param name="distortion">distortion to use</param>
        /// <param name="amount">Number of times to distort</param>
  /*      public static void Distort(DistortionStyle distortion, int amount, Bitmap im)
        {
            if (amount <= 0) return;
            for (int i = 0; i < amount; i++)
            {
                Distort(distortion, im);
             //   Conv3x3(im as Bitmap, distortion);
            }
        } */

        private static bool Conv3x3(Bitmap b, DistortionMethod m)
        {
            // Avoid divide by zero errors
            if (0 == m.Factor)
                return false; Bitmap

            // GDI+ still lies to us - the return format is BGR, NOT RGB. 
            bSrc = (Bitmap)b.Clone();
            BitmapData bmData = b.LockBits(new Rectangle(0, 0, b.Width, b.Height),
                                ImageLockMode.ReadWrite,
                                PixelFormat.Format24bppRgb);
            BitmapData bmSrc = bSrc.LockBits(new Rectangle(0, 0, bSrc.Width, bSrc.Height),
                               ImageLockMode.ReadWrite,
                               PixelFormat.Format24bppRgb);
            int stride = bmData.Stride;
            int stride2 = stride * 2;

            System.IntPtr Scan0 = bmData.Scan0;
            System.IntPtr SrcScan0 = bmSrc.Scan0;

            unsafe
            {
                byte* p = (byte*)(void*)Scan0;
                byte* pSrc = (byte*)(void*)SrcScan0;
                int nOffset = stride - b.Width * 3;
                int nWidth = b.Width - 2;
                int nHeight = b.Height - 2;

                int nPixel;

                for (int y = 0; y < nHeight; ++y)
                {
                    for (int x = 0; x < nWidth; ++x)
                    {
                        nPixel = ((((pSrc[2] * m.TopLeft) +
                            (pSrc[5] * m.TopMid) +
                            (pSrc[8] * m.TopRight) +
                            (pSrc[2 + stride] * m.MidLeft) +
                            (pSrc[5 + stride] * m.Pixel) +
                            (pSrc[8 + stride] * m.MidRight) +
                            (pSrc[2 + stride2] * m.BottomLeft) +
                            (pSrc[5 + stride2] * m.BottomMid) +
                            (pSrc[8 + stride2] * m.BottomRight))
                            / m.Factor) + m.Offset);

                        if (nPixel < 0) nPixel = 0;
                        if (nPixel > 255) nPixel = 255;
                        p[5 + stride] = (byte)nPixel;

                        nPixel = ((((pSrc[1] * m.TopLeft) +
                            (pSrc[4] * m.TopMid) +
                            (pSrc[7] * m.TopRight) +
                            (pSrc[1 + stride] * m.MidLeft) +
                            (pSrc[4 + stride] * m.Pixel) +
                            (pSrc[7 + stride] * m.MidRight) +
                            (pSrc[1 + stride2] * m.BottomLeft) +
                            (pSrc[4 + stride2] * m.BottomMid) +
                            (pSrc[7 + stride2] * m.BottomRight))
                            / m.Factor) + m.Offset);

                        if (nPixel < 0) nPixel = 0;
                        if (nPixel > 255) nPixel = 255;
                        p[4 + stride] = (byte)nPixel;

                        nPixel = ((((pSrc[0] * m.TopLeft) +
                                       (pSrc[3] * m.TopMid) +
                                       (pSrc[6] * m.TopRight) +
                                       (pSrc[0 + stride] * m.MidLeft) +
                                       (pSrc[3 + stride] * m.Pixel) +
                                       (pSrc[6 + stride] * m.MidRight) +
                                       (pSrc[0 + stride2] * m.BottomLeft) +
                                       (pSrc[3 + stride2] * m.BottomMid) +
                                       (pSrc[6 + stride2] * m.BottomRight))
                            / m.Factor) + m.Offset);

                        if (nPixel < 0) nPixel = 0;
                        if (nPixel > 255) nPixel = 255;
                        p[3 + stride] = (byte)nPixel;

                        p += 3;
                        pSrc += 3;
                    }

                    p += nOffset;
                    pSrc += nOffset;
                }
            }

            b.UnlockBits(bmData);
            bSrc.UnlockBits(bmSrc);
            return true;
        }


        private void DoNothing(Bitmap bm)
        {
            // The bitmap object will let us get at the actual pixel data
            System.Drawing.Imaging.BitmapData data = bm.LockBits(new Rectangle(0, 0, bm.Width, bm.Height),
                                ImageLockMode.ReadWrite,
                                PixelFormat.Format24bppRgb);
            unsafe
            {
                //Create a pointer into the structure
                byte* imgPtr = (byte*)(data.Scan0);
                // Now we will walk through the bitmap by row and column
                for (int i = 0; i < data.Height; i++)
                {
                    for (int j = 0; j < data.Width; j++)
                    {
                        // Here is a pixel
                        //Blue value:
                        byte blue = (byte)*imgPtr;
                        // advance one step to the green value
                        imgPtr++;
                        byte green = (byte)*imgPtr;
                        imgPtr++;
                        // advance one step to the red value
                        byte red = (byte)*imgPtr;
                        imgPtr++;
                    }
                    //Go to the next row
                    imgPtr += data.Stride - data.Width * 3;
                }
            }
            bm.UnlockBits(data);
        }
    }

    /// <summary>
    /// A class used for basic Image Manipulation.  Note that the System.Drwaing.Bitmap class 
    /// contains many useful functions.  The functions in this class do not damage the images passed to them 
    /// but instead returns values.
    /// </summary>
    public abstract class ImageGenerator
    {
        public static Image Distort(DistortionStyle ds, Image im)
        {
            Bitmap bm = new Bitmap(im.Clone() as Bitmap);
            ImageManipulator.Distort(ds, bm);
            return bm;
        }

        public static Image Dither(DitherStyle ds, Image im)
        {
            Bitmap bm = new Bitmap(im.Clone() as Bitmap);
            ImageManipulator.Dither(ds, bm);
            return bm;
        }

        public static Image Operation(OperationType ot, Image im1, Image im2)
        {
            Bitmap bm1 = new Bitmap(im1.Clone() as Bitmap);
            Bitmap bm2 = new Bitmap(im2.Clone() as Bitmap);
            ImageManipulator.Operation(ot, ref bm1, bm2);
            return bm1;
        }


        public static Image AlphaBlend(Image im1, Image im2, float blendAmount)
        {
            Bitmap bm1 = new Bitmap(im1.Clone() as Bitmap);
            Bitmap bm2 = new Bitmap(im2.Clone() as Bitmap);
            ImageManipulator.AlphaBlend(ref bm1, bm2, blendAmount);
            return bm1;
        }

        public static Image Resize(int newWidth, int newHeight, Image im1)
        {
            Bitmap bm1 = new Bitmap(im1.Clone() as Bitmap);
            ImageManipulator.Resize(newWidth, newHeight, ref bm1);
            return bm1;
        }

        public static Image FormattedByteArrayToImage(byte[] b)
        {
            return ImageManipulator.FormattedByteArrayToImage(b);
        }

        public static byte[] ImageToFormattedByteArray(ImageFormat imf, Image im)
        {
            return ImageManipulator.ImageToFormattedByteArray(imf, im);
        }
    }

    /// <summary>
    /// Provides Methods for analyzing images.
    /// </summary>
    public abstract class ImageAnalyzer
    {
        /// <summary>
        ///  Returns a value between 0 and 1 that representations the number of pixels in the image that have been changed.
        /// </summary>
        /// <param name="im1"></param>
        /// <param name="im2"></param>
        /// <returns></returns>
        public static float PercentDifferentPixels(int Threshold, Bitmap im1, Bitmap im2)
        {
            Bitmap bm = ImageGenerator.Operation(OperationType.Subtract, im1, im2) as Bitmap;
            if (null == bm) return 1;
            int totalChanged = 0;
            BitmapData data = bm.LockBits(new Rectangle(0, 0, bm.Width, bm.Height),
                    ImageLockMode.ReadWrite,
                    PixelFormat.Format24bppRgb);
            unsafe
            {
                byte* imgPtr = (byte*)(data.Scan0);
                for (int i = 0; i < data.Height; i++)
                {
                    for (int j = 0; j < data.Width; j++)
                    {
                        byte* otherPtr = imgPtr;
                        int b = (int)*otherPtr;
                        otherPtr++;
                        int g = (int)*otherPtr;
                        otherPtr++;
                        int r = (int)*otherPtr;
                        if ((b + g + r) > Threshold) totalChanged++;

                        imgPtr += 3;
                    }
                    imgPtr += data.Stride - data.Width * 3;
                }
            }
            bm.UnlockBits(data);
            return (float) ((float) totalChanged ) / ((float) (bm.Width * bm.Height));
        }
    

        public static float PercentDifferentDegree(Bitmap im1, Bitmap im2)
        {
            Bitmap bm = ImageGenerator.Operation(OperationType.Subtract, im1, im2) as Bitmap;
            if (null == bm) return 1;
            int totalChanged = 0;
            BitmapData data = bm.LockBits(new Rectangle(0, 0, bm.Width, bm.Height),
                    ImageLockMode.ReadWrite,
                    PixelFormat.Format24bppRgb);
            unsafe
            {
                byte* imgPtr = (byte*)(data.Scan0);
                for (int i = 0; i < data.Height; i++)
                {
                    for (int j = 0; j < data.Width; j++)
                    {
                        byte* otherPtr = imgPtr;
                        int b = (int)*otherPtr;
                        otherPtr++;
                        int g = (int)*otherPtr;
                        otherPtr++;
                        int r = (int)*otherPtr;
                        if ((b + g + r) > 0) totalChanged += (b+g+r);

                        imgPtr += 3;
                    }
                    imgPtr += data.Stride - data.Width * 3;
                }
            }
            bm.UnlockBits(data);
            return (float) ((float) totalChanged ) / ((float) ((bm.Width * bm.Height) *3*255));
        }
    }

   
}
