using EasyImages;
using GT.Net;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Net.Sockets;
using System.Text;

namespace MEdiaSpaceCore
{
    public class VideoWrapper
    {
        // Callback bits so the hosting UI can get a copy of images as they're sent out
        // This is set by the constructor and should *not* be null
        public delegate void LocalImageSentHandler(Image image);
        public event LocalImageSentHandler ImageSent = null;

        // Callback bits so the hosting UI can get a copy of images as they're received
        public delegate void RemoteImageReceivedHandler(Image image);
        public event RemoteImageReceivedHandler ImageReceived = null;

        // Status variable
        private bool _running = false;
        private bool _ishomeinstance = false;

        // Networking bits
        private IObjectStream _stream = null;

        // Camera bits
        private EasyImages.CameraClient _camera;
        private EasyImages.CameraServer _camserver;
        private long _previousFrameBroadcastTime;

        // Local cache of image capture and encoding parameters
        private Constants.VIDEO_PROTOCOLS _currentVideoProtocol;
        private double _currentFramerate;
        private int _currentJPEGCompression;

        // Stop-and-go protocol support
        private bool _ackReceived = false;


        //
        // Lifecycle
        //
        public VideoWrapper(Constants.VIDEO_PROTOCOLS videoProtocol,
                            double framerate,
                            int jpegCompression,
                            bool ishomeinstance,
                            LocalImageSentHandler imageSentHandler)
        {
            // But we should have an imageSentHandler (to see our own video frames)
            if (imageSentHandler == null)
                throw new ArgumentNullException("imageSentHandler");
            ImageSent += imageSentHandler;

            // And some valid imaging parameters
            SetVideoProtocol(videoProtocol);
            SetFramerate(framerate);
            SetJPEGCompression(jpegCompression);

            // Set up and start the camera server...
            _ishomeinstance = ishomeinstance;
            _camserver = new EasyImages.CameraServer();
            _camserver.CameraName = "DefaultCamera";
            _camserver.FrameSize = ishomeinstance ? Properties.Settings.Default.homecapturesize : Properties.Settings.Default.officecapturesize;
            _camserver.MaxFrameRate = ishomeinstance ? Properties.Settings.Default.homecapturerate : Properties.Settings.Default.officecapturerate;
            _camserver.Start();

            // And, finally, set up and start our cameras!
            _camera = new EasyImages.CameraClient("DefaultCamera");
            if (_camera == null)
                throw new Exception("Can't connect to the camera server");
            _camera.SpawnServer = false;
            _camera.ReceivedFrame += new CameraClient.ReceivedFrameEventHandler(camera_ReceivedFrame);
            _camera.FramesPerSecond = (int)Constants.MAX_FRAMERATE;
            _camera.Start();
        }


        //
        // Externally-visible methods
        //
        public void Start()
        {
            if (_running == true)
                return;

            // Reset our send timer
            _previousFrameBroadcastTime = 0;

            // Send one acknowledgement packet across the wire so we start receiving our images
            if (_stream != null)
                _stream.Send(new ACKPacket());
            
            _running = true;
        }

        public void Stop()
        {
            if (_running == false)
                return;

            _running = false;
        }
        
        public void SetStream(IObjectStream stream)
        {
            _stream = stream;

            if (_stream == null)
                return;

            if (_running == true)
                // Send one acknowledgement packet across the wire so we start receiving our images
                _stream.Send(new ACKPacket());
        }

        public void SetVideoProtocol(Constants.VIDEO_PROTOCOLS videoProtocol)
        {
            _currentVideoProtocol = videoProtocol;
        }

        public void SetFramerate(double framerate)
        {
            if (framerate <= 0.0 || framerate > Constants.MAX_FRAMERATE)
                throw new ArgumentException("Invalid framerate", "framerate");
            _currentFramerate = framerate;
        }

        public void SetJPEGCompression(int jpegCompression)
        {
            if (jpegCompression < 0 || jpegCompression > 100)
                throw new ArgumentException("Invalid compression ratio", "jpegCompression");
            _currentJPEGCompression = jpegCompression;
        }

        public void ResetCapture()
        {
            _camserver.FrameSize = _ishomeinstance ? Properties.Settings.Default.homecapturesize : Properties.Settings.Default.officecapturesize;
            _camserver.MaxFrameRate = _ishomeinstance ? Properties.Settings.Default.homecapturerate : Properties.Settings.Default.officecapturerate;

            _camserver.Stop();
            _camserver.Start();
        }

        public void SlowDown()
        {
            // Introduce some lag to clear the network backlog
            _previousFrameBroadcastTime = DateTime.Now.Ticks;
        }

        public void HandleIncomingMessage(Object receivedObject)
        {
            if (receivedObject is ACKPacket)
            {
                // Record receipt of the acknowledgement packet
                _ackReceived = true;
            }
            else if (receivedObject is VideoPacket)
            {
                VideoPacket packet = (VideoPacket)receivedObject;

                // Attempt to decode the video packet
                Bitmap bmap = packet.getBitmap();

                // If it worked out and we have somewhere to send it, then send it
                if (bmap != null && ImageReceived != null)
                    ImageReceived(bmap);

                // In any case, tell the sender that we're ready for the next frame
                if (_stream != null)
                    _stream.Send(new ACKPacket());
            }
            else
            {
                // Not sure what just came across the wire...
                throw new Exception("Unknown packet received!");
            }
        }

        
        //
        // EasyImages (camera) callback
        //
        private void camera_ReceivedFrame(object sender, CameraEventArgs e)
        {
            // Encode the frame first...
            VideoPacket packet = new VideoPacket(e.Bitmap, _currentVideoProtocol, _currentJPEGCompression);

            // Perform four checks before trying to send the video packet: running, connected,
            // framerate compliance, and ack received

            // First check: Are we running? If not, just bail
            if (_running == false)
            {
                return;
            }

            // Second check: Are we connected to a remote image receiver?
            // If not, just echo the video frame to our local image mirror and stall for a cycle
            if (_stream == null)
            {
                // The local image mirror should always be set, but check just in case...
                if (ImageSent != null)
                    ImageSent(e.Bitmap);

                return;
            }

            // Third check: has it been long enough since we sent the last frame?
            // If no, then just stall for a capture cycle
            double timeSinceLastBroadcast = (System.DateTime.Now.Ticks - _previousFrameBroadcastTime) / 10000000.0;
            if (timeSinceLastBroadcast > 0.0 && timeSinceLastBroadcast < (1.0 / _currentFramerate))
            {
                return;
            }

            // Fourth check: have we received our acknowledgement from the previous packet?
            // If no, then just stall for a capture cycle
            if (_ackReceived == false)
            {
                return;
            }

            // Write the packet to the network (with error checking)
            _stream.Send(packet, new MessageDeliveryRequirements(Reliability.Reliable, MessageAggregation.Immediate, Ordering.Ordered));

            // Record the time we sent the frame out
            _previousFrameBroadcastTime = DateTime.Now.Ticks;

            // Echo the frame back to the local terminal
            // (The local image mirror should always be set, but check just in case...)
            if (ImageSent != null)
                ImageSent(e.Bitmap);

            // And clear our acknowledgement flag (we're waiting for another one, now...)
            _ackReceived = false;
        }
    }
}
