using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Text;
using System.Threading;
using System.Timers;
using System.Windows.Forms;
using GT.Net;
using MEdiaSpaceCore;

namespace MEdiaSpaceHome
{
    partial class HomeForm : CustomForm
    {
        // A couple of constants to make the code easier to read
        private bool NOT_CONNECTED = false;

        // Private status variables
        private static bool _connected = false;
        protected bool _talking = false;

        // Shared variable manager
        Variables _variables = null;

        // Webcam source variable
        protected VideoWrapper _videoWrapper = null;

        // Audio wrapper variable
        protected AudioWrapper _audioWrapper = null;

        // Networking bits
        protected Client _client = null;
        protected IObjectStream _videoStream = null;
        protected IObjectStream _valueStream = null;
        protected ISessionStream _sessionStream = null;
        protected System.Timers.Timer _timer = null;
        protected IPAddress officeip = null;

        // UI status variables
        private bool _isDragging = false;
        private int _deltaX, _deltaY;
        private double _positionX, _positionY;

        //
        // Lifecycle
        //
        public HomeForm()
        {
            InitializeComponent();
        }

        private void SaveIPSetting()
        {
            // Save the IP address if successful
            Properties.Settings.Default.LastIP = officeip.ToString();

            // Add to the list of past IP addresses
            if (!Properties.Settings.Default.IPHistory.Contains("|" + officeip.ToString() + "|"))
            {
                if (Properties.Settings.Default.IPHistory == string.Empty) Properties.Settings.Default.IPHistory = "|";
                Properties.Settings.Default.IPHistory += officeip.ToString() + "|";
            }

            Properties.Settings.Default.Save();
        }


        // 
        // UI change handlers
        //
        private void HomeForm_Load(object sender, EventArgs e)
        {
            // Connect to our variable cache and set up our callbacks
            _variables = new Variables();
            _variables.DoorActuationAvailable += new Variables.DoorActuationAvailableHandler(_variables_DoorActuationAvailable);
            _variables.DoorbellStatusChanged += new Variables.DoorbellStatusChangeHandler(_variables_DoorbellStatusChanged);
            _variables.HomeMessageChanged += new Variables.MessageChangeHandler(_variables_HomeMessageChanged);
            _variables.HomeVideoConfigurationChanged += new Variables.VideoConfigurationChangeHandler(_variables_HomeVideoConfigurationChanged);
            _variables.OfficeMessageChanged += new Variables.MessageChangeHandler(_variables_OfficeMessageChanged);
            _variables.OfficeSkypeHandleChanged += new Variables.SkypeHandleChangeHandler(_variables_OfficeSkypeHandleChanged);

            // Video feed setup
            AddChatMessage("MEdia Space", "Setting up video...");
            _videoWrapper = new VideoWrapper(_variables.homeProtocol,
                                             _variables.homeFrameRateSlow,
                                             _variables.homeCompression,
                                             true,
                                             new VideoWrapper.LocalImageSentHandler(SetLocalImage));
            _videoWrapper.ImageReceived += new VideoWrapper.RemoteImageReceivedHandler(SetRemoteImage);
            _videoWrapper.Start();

            // Audio feed setup
            AddChatMessage("MEdia Space", "Connecting to Skype application...");
            _audioWrapper = new AudioWrapper(null);
            _variables.homeSkypeHandle = _audioWrapper.CurrentHandle();
            _audioWrapper.CallConnected += new AudioWrapper.CallConnectedHandler(audioWrapper_CallConnected);
            _audioWrapper.CallingAvailable += new AudioWrapper.CallingAvailableHandler(SetPhoneButtonEnabled);

            // UI setup
            SetConnected(false);
            SetDoorButtonEnabled(false);
            SetPhoneButtonStatus(NOT_CONNECTED);
            SetDocumentButtonEnabled(false);
            SetToolsButtonEnabled(false);
            SetLocalImage(null);
            SetRemoteImage(null);

            _positionX = (double)(outgoingPictureBox.Location.X - incomingPictureBox.Location.X) /
                         (double)incomingPictureBox.Width;
            _positionY = (double)(outgoingPictureBox.Location.Y - incomingPictureBox.Location.Y) /
                         (double)incomingPictureBox.Height;

            // Automatically start to connect if the option is turned on!
            if (_variables.homeConnectOnStartup)
                connectButton_Click(null, null);
        }

        void HomeForm_FormClosed(object sender, System.Windows.Forms.FormClosedEventArgs e)
        {
            // Close the audio connection (if there is one)
            if (_audioWrapper != null && _audioWrapper.IsTalking())
                _audioWrapper.StopCall();

            // Shut off the video
            if (_videoWrapper != null)
            {
                _videoWrapper.Stop();
                _videoWrapper.SetStream(null);
            }

            // Wipe our identity for the audio and shut off the network
            if (_client != null)
            {
                _variables.homeSkypeHandle = null;
                _client.Stop();
            }

            // And exit
            Application.Exit();
        }

        void HomeForm_Resize(object sender, System.EventArgs e)
        {
            // Support for minimizing to the tray icon
            if (WindowState == FormWindowState.Minimized)
            {
                notifyIcon.Visible = true;
                Hide();
                return;
            }
            else
            {
                notifyIcon.Visible = false;
            }

            // Set a hard stop for all the madness!
            MinimumSize = new Size(Constants.HOME_INVIDEO_MINIMUM_WIDTH + Constants.HOME_INVIDEO_HORIZONTAL_GUTTER,
                                   Constants.HOME_INVIDEO_MINIMUM_HEIGHT + Constants.HOME_INVIDEO_VERTICAL_GUTTER);

            // Toggle the visibility (and size) of the chat boxes based on the size of the incoming pbox
            if (chatOutputBox.Width < Constants.HOME_CHAT_MIN_WIDTH_VISIBLE ||
                Height < Constants.HOME_SECOND_CONTRACT_THRESHOLD)
            {
                chatInputBox.Visible = false;
                chatOutputBox.Visible = false;
            }
            else if (Height < Constants.HOME_FIRST_CONTRACT_THRESHOLD)
            {
                chatInputBox.Visible = false;
                chatOutputBox.Visible = true;
                chatOutputBox.Height = Constants.HOME_OUTCHAT_HALF_HEIGHT;
                chatOutputBox.Top = Height - Constants.HOME_CONTRACTED_CONTROL_DIST_TO_BOTTOM;
            }
            else
            {
                chatInputBox.Visible = true;
                chatOutputBox.Visible = true;
                chatOutputBox.Height = Constants.HOME_OUTCHAT_FULL_HEIGHT;
                chatOutputBox.Top = Height - Constants.HOME_EXPANDED_CONTROL_DIST_TO_BOTTOM;
            }

            // Toggle the visiblility of the buttons based on the size of the incoming pbox
            if (Height < Constants.HOME_SECOND_CONTRACT_THRESHOLD)
            {
                connectButton.Visible = false;
                doorButton.Visible = false;
                phoneButton.Visible = false;

                documentButton.Visible = false;
                toolsButton.Visible = false;
            }
            else if (Height < Constants.HOME_FIRST_CONTRACT_THRESHOLD)
            {
                connectButton.Visible = true;
                connectButton.Top = Height - Constants.HOME_CONTRACTED_CONTROL_DIST_TO_BOTTOM;
                doorButton.Visible = true;
                doorButton.Top = Height - Constants.HOME_CONTRACTED_CONTROL_DIST_TO_BOTTOM;
                phoneButton.Visible = true;
                phoneButton.Top = Height - Constants.HOME_CONTRACTED_CONTROL_DIST_TO_BOTTOM;

                documentButton.Visible = false;
                toolsButton.Visible = false;
            }
            else
            {
                connectButton.Visible = true;
                connectButton.Top = Height - Constants.HOME_EXPANDED_CONTROL_DIST_TO_BOTTOM;
                doorButton.Visible = true;
                doorButton.Top = Height - Constants.HOME_EXPANDED_CONTROL_DIST_TO_BOTTOM;
                phoneButton.Visible = true;
                phoneButton.Top = Height - Constants.HOME_EXPANDED_CONTROL_DIST_TO_BOTTOM;

                documentButton.Visible = true;
                toolsButton.Visible = true;
            }

            // Scale the height of the incomingPictureBox
            int computedHeight = Height - Constants.HOME_INVIDEO_DIST_TO_BOTTOM;
            if (documentButton.Visible == false)
                computedHeight += Constants.HOME_SECONDBUTTONS_HEIGHT;
            if (connectButton.Visible == false)
                computedHeight += Constants.HOME_FIRSTBUTTONS_HEIGHT;
            incomingPictureBox.Height = computedHeight;

            // Scale the outgoingPictureBox to stay the same size relative to the incomingPictureBox
            // and toggle its visibility based on the size of its "parent"
            if (incomingPictureBox.Width >= Constants.HOME_OUTVIDEO_MIN_WIDTH_VISIBLE ||
                incomingPictureBox.Height >= Constants.HOME_OUTVIDEO_MIN_HEIGHT_VISIBLE)
            {
                outgoingPictureBox.Visible = true;

                int previousWidth = outgoingPictureBox.Width;
                int previousHeight = outgoingPictureBox.Height;
                int newWidth = incomingPictureBox.Width *
                               Constants.HOME_OUTVIDEO_INITIAL_WIDTH /
                               Constants.HOME_INVIDEO_INITIAL_WIDTH;
                int newHeight = incomingPictureBox.Height *
                                Constants.HOME_OUTVIDEO_INITIAL_HEIGHT /
                                Constants.HOME_INVIDEO_INITIAL_HEIGHT;
                int newX = (int)(_positionX * (double)incomingPictureBox.Width) + incomingPictureBox.Location.X;
                newX = Math.Max(newX, incomingPictureBox.Location.X + Constants.PBOX_MARGIN);
                newX = Math.Min(newX, incomingPictureBox.Location.X + incomingPictureBox.Width - Constants.PBOX_MARGIN);
                int newY = (int)(_positionY * (double)incomingPictureBox.Height) + incomingPictureBox.Location.Y;
                newY = Math.Max(newY, incomingPictureBox.Location.Y + Constants.PBOX_MARGIN);
                newY = Math.Min(newY, incomingPictureBox.Location.Y + incomingPictureBox.Height - Constants.PBOX_MARGIN);

                outgoingPictureBox.Location = new Point(newX, newY);
                outgoingPictureBox.Size = new Size(newWidth, newHeight);
            }
            else
            {
                outgoingPictureBox.Visible = false;
            }
        }

        private void connectButton_Click(object sender, EventArgs e)
        {
            if (!_connected)
            {
                // Call up connection dialog.
                ConnectDialog frm = new ConnectDialog();
                if (frm.ShowDialog() != DialogResult.OK)
                {
                    return;
                }

                officeip = frm.SelectedIP;
                AddChatMessage("MEdia Space", "Connecting to the office node at " + officeip.ToString() + "...");

                _client = new Client(new MEdiaSpaceClientConfiguration());
                _client.ErrorEvent += new ErrorEventNotication(_client_ErrorEvent);
                _client.Start();
                ChannelDeliveryRequirements cdr = new ChannelDeliveryRequirements(Reliability.Reliable, MessageAggregation.Immediate, Ordering.Ordered);
                try
                {
                    _sessionStream = _client.GetSessionStream(officeip.ToString(), Convert.ToString(Constants.OFFICE_PORT), (byte)Constants.SESSION_CHANNEL, cdr);
                    SaveIPSetting();
                }
                catch (CannotConnectException)
                {
                    AddChatMessage("MEdia Space", "Cannot connect to the office node. (The office node might not be running right now...)");

                    if (_client != null)
                    {
                        _client.Stop();
                        _client.Dispose();
                        _client = null;
                    }

                    return;
                }
                _videoStream = _client.GetObjectStream(officeip.ToString(), Convert.ToString(Constants.OFFICE_PORT), (byte)Constants.VIDEO_CHANNEL, cdr);
                _videoWrapper.SetStream(_videoStream);
                _valueStream = _client.GetObjectStream(officeip.ToString(), Convert.ToString(Constants.OFFICE_PORT), (byte)Constants.DATA_CHANNEL, cdr);
                _variables.SetStream(_valueStream);

                _timer = new System.Timers.Timer(Constants.NETWORK_CHECK_RATE);
                _timer.Elapsed += new ElapsedEventHandler(_timer_Elapsed);
                _timer.Start();
            }
            else
            {
                // Close the audio connection (if there is one)
                if (_audioWrapper != null && _audioWrapper.IsTalking())
                    _audioWrapper.StopCall();

                // Wipe our identity
                _variables.homeSkypeHandle = null;

                // Shut off the video
                _videoWrapper.Stop();
                _videoWrapper.SetStream(null);
            }
        }

        private void doorButton_Click(object sender, EventArgs e)
        {
            _variables.ActuateDoor();
        }

        void outgoingPictureBox_MouseUp(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            _isDragging = false;

            _positionX = (double)(outgoingPictureBox.Location.X - incomingPictureBox.Location.X) /
                         (double)incomingPictureBox.Width;
            _positionY = (double)(outgoingPictureBox.Location.Y - incomingPictureBox.Location.Y) /
                         (double)incomingPictureBox.Height;
        }

        void outgoingPictureBox_MouseMove(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            if (_isDragging)
            {
                int newTop = outgoingPictureBox.Top + (e.Y - _deltaY);
                int newLeft = outgoingPictureBox.Left + (e.X - _deltaX);

                newTop = Math.Max(newTop, incomingPictureBox.Top + Constants.PBOX_MARGIN);
                newTop = Math.Min(newTop, incomingPictureBox.Bottom - Constants.PBOX_MARGIN - outgoingPictureBox.Height);

                newLeft = Math.Max(newLeft, incomingPictureBox.Left + Constants.PBOX_MARGIN);
                newLeft = Math.Min(newLeft, incomingPictureBox.Right - Constants.PBOX_MARGIN - outgoingPictureBox.Width);

                outgoingPictureBox.Top = newTop;
                outgoingPictureBox.Left = newLeft;
            }
        }

        void outgoingPictureBox_MouseDown(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            _isDragging = true;

            _deltaX = e.X;
            _deltaY = e.Y;
        }

        private void chatInputBox_KeyUp(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
            {
                _variables.homeMessage = chatInputBox.Text;
                chatInputBox.Text = "";
            }
        }

        private void documentButton_Click(object sender, EventArgs e)
        {
            // TODO
        }

        private void phoneButton_Click(object sender, EventArgs e)
        {
            if (!_talking && _audioWrapper.CurrentHandle() != null)
            {
                Thread audioSourceThread = new Thread(new ThreadStart(_audioWrapper.StartCall));
                audioSourceThread.IsBackground = true;
                audioSourceThread.Start();
            }
            else
                _audioWrapper.StopCall();
        }

        private void toolsButton_Click(object sender, EventArgs e)
        {
            ControlPanel cp = new ControlPanel();
            if (DialogResult.OK == cp.ShowDialog(this))
            {
                _videoWrapper.ResetCapture();
            }
        }

        private void notifyIcon_DoubleClick(object sender, System.EventArgs e)
        {
            this.Show();
            WindowState = FormWindowState.Normal;
            notifyIcon.Visible = false;
        }


        //
        // Networking handlers
        //
        private void _timer_Elapsed(object sender, ElapsedEventArgs e)
        {
            // If the network has died, stop trying to read from it
            if (_client == null || _client.Active == false)
            {
                if (_timer != null)
                {
                    _timer.Stop();
                    _timer.Dispose();
                    _timer = null;
                }

                return;
            }
            
            _client.Update();

            // Handle session messages first
            SessionMessage sm;
            while ((sm = _sessionStream.DequeueMessage(0)) != null)
                HandleIncomingSessionMessage(sm);

            // Handle video-related messages next
            Object packet;
            while ((packet = _videoStream.DequeueMessage(0)) != null)
                _videoWrapper.HandleIncomingMessage(packet);

            // Handle data messages last
            while ((packet = _valueStream.DequeueMessage(0)) != null)
                _variables.HandleIncomingMessage(packet);
        }


        private void HandleIncomingSessionMessage(SessionMessage message)
        {
            if (message.Action == SessionAction.Joined)
            {
                // Update our status for the host
                _variables.homeSkypeHandle = _audioWrapper.CurrentHandle();
                _variables.homeStatus = Constants.HOME_CONNECTED_STRING;

                AddChatMessage("MEdia Space", "Connected!");

                _connected = true;
                SetConnected(true);
                SetPhoneButtonEnabled(true);
                SetPhoneButtonStatus(NOT_CONNECTED);
                SetDocumentButtonEnabled(true);
                SetToolsButtonEnabled(true);
                SetRemoteImage(null);
            }
            if (message.Action == SessionAction.Left || message.Action == SessionAction.Inactive)
            {
                _connected = false;
                SetConnected(false);
                SetPhoneButtonEnabled(false);
                SetPhoneButtonStatus(NOT_CONNECTED);
                SetDoorButtonEnabled(false);
                SetDocumentButtonEnabled(false);
                SetToolsButtonEnabled(false);
                SetLocalImage(null);
                SetRemoteImage(null);

                // Close the audio connection (if there is one)
                if (_audioWrapper != null && _audioWrapper.IsTalking())
                    _audioWrapper.StopCall();

                // Close the video connection (if there is one)
                _videoWrapper.SetStream(null);

                AddChatMessage("MEdia Space", "Disconnected from the office node");
            }
        }

        void _client_ErrorEvent(ErrorSummary summary)
        {
            switch (summary.ErrorCode) {
                case SummaryErrorCode.InvalidIncomingMessage:
                case SummaryErrorCode.MessagesCannotBeSent:
                    break;

                case SummaryErrorCode.RemoteUnavailable:
                case SummaryErrorCode.UserException:
                    AddChatMessage("MEdia Space", "Experiencing network difficulties...");

                    // Forcibly disconnect
                    HandleIncomingSessionMessage(new SessionMessage(0, 0, SessionAction.Left));
                    break;

                case SummaryErrorCode.TransportBacklogged:
                    _videoWrapper.SlowDown();
                    break;
            }
        }


        //
        // Audio wrapper change handler
        //
        private void audioWrapper_CallConnected(bool talking)
        {
            _talking = talking;
            SetPhoneButtonStatus(_talking);

            if (_talking)
                _videoWrapper.SetFramerate(_variables.homeFrameRateFast);
            else
                _videoWrapper.SetFramerate(_variables.homeFrameRateSlow);
        }


        //
        // Shared variable change handlers
        //
        void _variables_DoorActuationAvailable(bool available)
        {
            SetDoorButtonEnabled(available);
        }

        void _variables_DoorbellStatusChanged(bool status)
        {
            // TODO
        }

        void _variables_HomeMessageChanged(string message)
        {
            if (message != null)
                AddChatMessage("Home", message);
        }

        void _variables_HomeVideoConfigurationChanged(Constants.VIDEO_PROTOCOLS protocol, int compression, double frameRateSlow, double frameRateFast)
        {
            if (_videoWrapper != null) {
                _videoWrapper.SetVideoProtocol(protocol);
                _videoWrapper.SetJPEGCompression(compression);
                if (_talking)
                    _videoWrapper.SetFramerate(frameRateFast);
                else
                    _videoWrapper.SetFramerate(frameRateSlow);
            }
        }

        void _variables_OfficeMessageChanged(String message)
        {
            if (message != null)
                AddChatMessage("Office", message);
        }

        void _variables_OfficeSkypeHandleChanged(String skypeHandle)
        {
            _audioWrapper.SetSinkUsername(skypeHandle);

            if (skypeHandle != null)
                SetPhoneButtonEnabled(true);
            else
                SetPhoneButtonEnabled(false);
        }


        //
        // Content distribution functions
        //
        public void SetConnected(bool connected)
        {
            if (connected)
                connectButton.Image = global::MEdiaSpaceHome.Properties.Resources.disconnect;
            else
                connectButton.Image = global::MEdiaSpaceHome.Properties.Resources.connect;
        }

        public void SetDoorButtonEnabled(bool enabled)
        {
            // Have to use delegate magic since this can be a cross-thread call
            EnableControl(doorButton, enabled);
        }

        public void SetDocumentButtonEnabled(bool enabled)
        {
            // Have to use delegate magic since this can be a cross-thread call
            EnableControl(documentButton, enabled);
        }

        public void SetToolsButtonEnabled(bool enabled)
        {
            // Have to use delegate magic since this can be a cross-thread call
            EnableControl(toolsButton, enabled);
        }

        public void SetPhoneButtonEnabled(bool enabled)
        {
            // Have to use delegate magic since this can be a cross-thread call
            EnableControl(phoneButton, enabled);
        }

        public void SetPhoneButtonStatus(bool talking)
        {
            if (talking)
                phoneButton.Image = global::MEdiaSpaceHome.Properties.Resources.redphone;
            else
                phoneButton.Image = global::MEdiaSpaceHome.Properties.Resources.greenphone;
        }

        public void SetMotionDetection(int value, DateTime timestamp)
        {
            // TODO
        }

        public void SetRemoteImage(Image image)
        {
            // Don't bother if we're minimized...
            if (WindowState == FormWindowState.Minimized)
                return;

            DisplayImageInPictureBox(incomingPictureBox, image);
        }

        public void SetLocalImage(Image image)
        {
            // Don't bother if we're minimized...
            if (WindowState == FormWindowState.Minimized)
                return;

            if (null != image)
                image.RotateFlip(RotateFlipType.Rotate180FlipY);

            DisplayImageInPictureBox(outgoingPictureBox, image);
        }

        public void AddChatMessage(String sender, String message)
        {
            if (WindowState == FormWindowState.Minimized)
                notifyIcon.ShowBalloonTip(5000, sender, message, ToolTipIcon.Info);

            DisplayTextInControl(chatOutputBox, chatOutputBox.Text + sender + ": " + message + System.Environment.NewLine);
            FocusOnControl(chatInputBox);
        }
        
        public void ClearChatSpool()
        {
            DisplayTextInControl(chatOutputBox, "");
        }
    }
}